@@ -1,25 +1,28 @@
 /*
  * Copyright © 2013 Intel Corporation
  *
- * Permission to use, copy, modify, distribute, and sell this software and its
- * documentation for any purpose is hereby granted without fee, provided that
- * the above copyright notice appear in all copies and that both that copyright
- * notice and this permission notice appear in supporting documentation, and
- * that the name of the copyright holders not be used in advertising or
- * publicity pertaining to distribution of the software without specific,
- * written prior permission.  The copyright holders make no representations
- * about the suitability of this software for any purpose.  It is provided "as
- * is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining a
+ * copy of this software and associated documentation files (the "Software"),
+ * to deal in the Software without restriction, including without limitation
+ * the rights to use, copy, modify, merge, publish, distribute, sublicense,
+ * and/or sell copies of the Software, and to permit persons to whom the
+ * Software is furnished to do so, subject to the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
- * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
- * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
- * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
- * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
- * OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the next
+ * paragraph) shall be included in all copies or substantial portions of the
+ * Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
+ * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+ * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
+ * DEALINGS IN THE SOFTWARE.
  */
 
+#include "config.h"
+
 #include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
@@ -43,19 +46,12 @@
 #define WL_HIDE_DEPRECATED
 #include <wayland-server.h>
 
+#include "shared/helpers.h"
+#include "shared/xalloc.h"
 #include "window.h"
 
-#define MIN(x,y) (((x) < (y)) ? (x) : (y))
+#include "weston-egl-ext.h"
 
-#ifndef EGL_WL_create_wayland_buffer_from_image
-#define EGL_WL_create_wayland_buffer_from_image 1
-
-#ifdef EGL_EGLEXT_PROTOTYPES
-EGLAPI struct wl_buffer * EGLAPIENTRY eglCreateWaylandBufferFromImageWL(EGLDisplay dpy, EGLImageKHR image);
-#endif
-typedef struct wl_buffer * (EGLAPIENTRYP PFNEGLCREATEWAYLANDBUFFERFROMIMAGEWL) (EGLDisplay dpy, EGLImageKHR image);
-
-#endif
 
 static int option_blit;
 
@@ -548,13 +544,6 @@ surface_set_input_region(struct wl_client *client,
 	fprintf(stderr, "surface_set_input_region\n");
 }
 
-static void
-empty_region(pixman_region32_t *region)
-{
-	pixman_region32_fini(region);
-	pixman_region32_init(region);
-}
-
 static void
 surface_commit(struct wl_client *client, struct wl_resource *resource)
 {
@@ -572,7 +561,7 @@ surface_commit(struct wl_client *client, struct wl_resource *resource)
 	surface->pending.newly_attached = 0;
 
 	/* wl_surface.damage */
-	empty_region(&surface->pending.damage);
+	pixman_region32_clear(&surface->pending.damage);
 
 	/* wl_surface.frame */
 	wl_list_insert_list(&surface->frame_callback_list,
@@ -620,7 +609,7 @@ compositor_create_surface(struct wl_client *client,
 {
 	struct nested *nested = wl_resource_get_user_data(resource);
 	struct nested_surface *surface;
-	
+
 	surface = zalloc(sizeof *surface);
 	if (surface == NULL) {
 		wl_resource_post_no_memory(resource);
@@ -759,7 +748,7 @@ nested_init_compositor(struct nested *nested)
 
 	nested->egl_display = display_get_egl_display(nested->display);
 	extensions = eglQueryString(nested->egl_display, EGL_EXTENSIONS);
-	if (strstr(extensions, "EGL_WL_bind_wayland_display") == NULL) {
+	if (weston_check_egl_extension(extensions, "EGL_WL_bind_wayland_display") == NULL) {
 		fprintf(stderr, "no EGL_WL_bind_wayland_display extension\n");
 		return -1;
 	}
@@ -782,7 +771,7 @@ nested_init_compositor(struct nested *nested)
 		const char *func = "eglCreateWaylandBufferFromImageWL";
 		const char *ext = "EGL_WL_create_wayland_buffer_from_image";
 
-		if (strstr(extensions, ext)) {
+		if (weston_check_egl_extension(extensions, ext)) {
 			create_wayland_buffer_from_image =
 				(void *) eglGetProcAddress(func);
 			use_ss_renderer = 1;
@@ -976,6 +965,8 @@ ss_surface_init(struct nested_surface *surface)
 				      nested,
 				      SUBSURFACE_SYNCHRONIZED);
 
+	widget_set_use_cairo(ss_surface->widget, 0);
+
 	ss_surface->surface = widget_get_wl_surface(ss_surface->widget);
 	ss_surface->subsurface = widget_get_wl_subsurface(ss_surface->widget);
 
@@ -1118,8 +1109,11 @@ main(int argc, char *argv[])
 	struct display *display;
 	struct nested *nested;
 
-	parse_options(nested_options,
-		      ARRAY_LENGTH(nested_options), &argc, argv);
+	if (parse_options(nested_options,
+			  ARRAY_LENGTH(nested_options), &argc, argv) > 1) {
+		printf("Usage: %s [OPTIONS]\n  --blit or -b\n", argv[0]);
+		exit(1);
+	}
 
 	display = display_create(&argc, argv);
 	if (display == NULL) {
