@@ -2,27 +2,29 @@
  * Copyright © 2011 Benjamin Franzke
  * Copyright © 2011 Intel Corporation
  *
- * Permission to use, copy, modify, distribute, and sell this software and its
- * documentation for any purpose is hereby granted without fee, provided that
- * the above copyright notice appear in all copies and that both that copyright
- * notice and this permission notice appear in supporting documentation, and
- * that the name of the copyright holders not be used in advertising or
- * publicity pertaining to distribution of the software without specific,
- * written prior permission.  The copyright holders make no representations
- * about the suitability of this software for any purpose.  It is provided "as
- * is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining a
+ * copy of this software and associated documentation files (the "Software"),
+ * to deal in the Software without restriction, including without limitation
+ * the rights to use, copy, modify, merge, publish, distribute, sublicense,
+ * and/or sell copies of the Software, and to permit persons to whom the
+ * Software is furnished to do so, subject to the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
- * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
- * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
- * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
- * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
- * OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the next
+ * paragraph) shall be included in all copies or substantial portions of the
+ * Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
+ * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+ * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
+ * DEALINGS IN THE SOFTWARE.
  */
 
-#include <config.h>
+#include "config.h"
 
+#include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -32,9 +34,14 @@
 #include <sys/mman.h>
 
 #include <wayland-client.h>
-#include "../shared/os-compatibility.h"
+#include "shared/helpers.h"
+#include "shared/os-compatibility.h"
 
-#define ARRAY_LENGTH(a) (sizeof (a) / sizeof (a)[0])
+struct seat {
+	struct touch *touch;
+	struct wl_seat *seat;
+	struct wl_touch *wl_touch;
+};
 
 struct touch {
 	struct wl_display *display;
@@ -42,8 +49,6 @@ struct touch {
 	struct wl_compositor *compositor;
 	struct wl_shell *shell;
 	struct wl_shm *shm;
-	struct wl_seat *seat;
-	struct wl_touch *wl_touch;
 	struct wl_pointer *pointer;
 	struct wl_keyboard *keyboard;
 	struct wl_surface *surface;
@@ -199,18 +204,19 @@ static const struct wl_touch_listener touch_listener = {
 };
 
 static void
-seat_handle_capabilities(void *data, struct wl_seat *seat,
+seat_handle_capabilities(void *data, struct wl_seat *wl_seat,
 			 enum wl_seat_capability caps)
 {
-	struct touch *touch = data;
+	struct seat *seat = data;
+	struct touch *touch = seat->touch;
 
-	if ((caps & WL_SEAT_CAPABILITY_TOUCH) && !touch->wl_touch) {
-		touch->wl_touch = wl_seat_get_touch(seat);
-		wl_touch_set_user_data(touch->wl_touch, touch);
-		wl_touch_add_listener(touch->wl_touch, &touch_listener, touch);
-	} else if (!(caps & WL_SEAT_CAPABILITY_TOUCH) && touch->wl_touch) {
-		wl_touch_destroy(touch->wl_touch);
-		touch->wl_touch = NULL;
+	if ((caps & WL_SEAT_CAPABILITY_TOUCH) && !seat->wl_touch) {
+		seat->wl_touch = wl_seat_get_touch(wl_seat);
+		wl_touch_set_user_data(seat->wl_touch, touch);
+		wl_touch_add_listener(seat->wl_touch, &touch_listener, touch);
+	} else if (!(caps & WL_SEAT_CAPABILITY_TOUCH) && seat->wl_touch) {
+		wl_touch_destroy(seat->wl_touch);
+		seat->wl_touch = NULL;
 	}
 }
 
@@ -218,6 +224,21 @@ static const struct wl_seat_listener seat_listener = {
 	seat_handle_capabilities,
 };
 
+static void
+add_seat(struct touch *touch, uint32_t name, uint32_t version)
+{
+	struct seat *seat;
+
+	seat = malloc(sizeof *seat);
+	assert(seat);
+
+	seat->touch = touch;
+	seat->wl_touch = NULL;
+	seat->seat = wl_registry_bind(touch->registry, name,
+				      &wl_seat_interface, 1);
+	wl_seat_add_listener(seat->seat, &seat_listener, seat);
+}
+
 static void
 handle_ping(void *data, struct wl_shell_surface *shell_surface,
 	    uint32_t serial)
@@ -261,9 +282,7 @@ handle_global(void *data, struct wl_registry *registry,
 					      &wl_shm_interface, 1);
 		wl_shm_add_listener(touch->shm, &shm_listener, touch);
 	} else if (strcmp(interface, "wl_seat") == 0) {
-		touch->seat = wl_registry_bind(registry, name,
-					       &wl_seat_interface, 1);
-		wl_seat_add_listener(touch->seat, &seat_listener, touch);
+		add_seat(touch, name, version);
 	}
 }
 
