@@ -1,26 +1,27 @@
 /*
  * Copyright © 2008 Kristian Høgsberg
  *
- * Permission to use, copy, modify, distribute, and sell this software and its
- * documentation for any purpose is hereby granted without fee, provided that
- * the above copyright notice appear in all copies and that both that copyright
- * notice and this permission notice appear in supporting documentation, and
- * that the name of the copyright holders not be used in advertising or
- * publicity pertaining to distribution of the software without specific,
- * written prior permission.  The copyright holders make no representations
- * about the suitability of this software for any purpose.  It is provided "as
- * is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining a
+ * copy of this software and associated documentation files (the "Software"),
+ * to deal in the Software without restriction, including without limitation
+ * the rights to use, copy, modify, merge, publish, distribute, sublicense,
+ * and/or sell copies of the Software, and to permit persons to whom the
+ * Software is furnished to do so, subject to the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
- * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
- * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
- * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
- * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
- * OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the next
+ * paragraph) shall be included in all copies or substantial portions of the
+ * Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
+ * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
+ * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
+ * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
+ * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
+ * DEALINGS IN THE SOFTWARE.
  */
 
-#include <config.h>
+#include "config.h"
 
 #include <stdbool.h>
 #include <stdint.h>
@@ -42,7 +43,9 @@
 
 #include <wayland-client.h>
 
-#include "../shared/config-parser.h"
+#include "shared/config-parser.h"
+#include "shared/helpers.h"
+#include "shared/xalloc.h"
 #include "window.h"
 
 static int option_fullscreen;
@@ -121,6 +124,7 @@ init_state_machine(struct utf8_state_machine *machine)
 	machine->state = utf8state_start;
 	machine->len = 0;
 	machine->s.ch = 0;
+	machine->unicode = 0;
 }
 
 static enum utf8_state
@@ -132,28 +136,28 @@ utf8_next_char(struct utf8_state_machine *machine, unsigned char c)
 	case utf8state_reject:
 		machine->s.ch = 0;
 		machine->len = 0;
-		if(c == 0xC0 || c == 0xC1) {
+		if (c == 0xC0 || c == 0xC1) {
 			/* overlong encoding, reject */
 			machine->state = utf8state_reject;
-		} else if((c & 0x80) == 0) {
+		} else if ((c & 0x80) == 0) {
 			/* single byte, accept */
 			machine->s.byte[machine->len++] = c;
 			machine->state = utf8state_accept;
 			machine->unicode = c;
-		} else if((c & 0xC0) == 0x80) {
+		} else if ((c & 0xC0) == 0x80) {
 			/* parser out of sync, ignore byte */
 			machine->state = utf8state_start;
-		} else if((c & 0xE0) == 0xC0) {
+		} else if ((c & 0xE0) == 0xC0) {
 			/* start of two byte sequence */
 			machine->s.byte[machine->len++] = c;
 			machine->state = utf8state_expect1;
 			machine->unicode = c & 0x1f;
-		} else if((c & 0xF0) == 0xE0) {
+		} else if ((c & 0xF0) == 0xE0) {
 			/* start of three byte sequence */
 			machine->s.byte[machine->len++] = c;
 			machine->state = utf8state_expect2;
 			machine->unicode = c & 0x0f;
-		} else if((c & 0xF8) == 0xF0) {
+		} else if ((c & 0xF8) == 0xF0) {
 			/* start of four byte sequence */
 			machine->s.byte[machine->len++] = c;
 			machine->state = utf8state_expect3;
@@ -166,7 +170,7 @@ utf8_next_char(struct utf8_state_machine *machine, unsigned char c)
 	case utf8state_expect3:
 		machine->s.byte[machine->len++] = c;
 		machine->unicode = (machine->unicode << 6) | (c & 0x3f);
-		if((c & 0xC0) == 0x80) {
+		if ((c & 0xC0) == 0x80) {
 			/* all good, continue */
 			machine->state = utf8state_expect2;
 		} else {
@@ -177,7 +181,7 @@ utf8_next_char(struct utf8_state_machine *machine, unsigned char c)
 	case utf8state_expect2:
 		machine->s.byte[machine->len++] = c;
 		machine->unicode = (machine->unicode << 6) | (c & 0x3f);
-		if((c & 0xC0) == 0x80) {
+		if ((c & 0xC0) == 0x80) {
 			/* all good, continue */
 			machine->state = utf8state_expect1;
 		} else {
@@ -188,7 +192,7 @@ utf8_next_char(struct utf8_state_machine *machine, unsigned char c)
 	case utf8state_expect1:
 		machine->s.byte[machine->len++] = c;
 		machine->unicode = (machine->unicode << 6) | (c & 0x3f);
-		if((c & 0xC0) == 0x80) {
+		if ((c & 0xC0) == 0x80) {
 			/* all good, accept */
 			machine->state = utf8state_accept;
 		} else {
@@ -200,7 +204,7 @@ utf8_next_char(struct utf8_state_machine *machine, unsigned char c)
 		machine->state = utf8state_reject;
 		break;
 	}
-	
+
 	return machine->state;
 }
 
@@ -284,7 +288,7 @@ static void
 apply_char_set(character_set cs, union utf8_char *utf8)
 {
 	int i = 0;
-	
+
 	while (cs[i].match.byte[0]) {
 		if ((*utf8).ch == cs[i].match.ch) {
 			*utf8 = cs[i].replace;
@@ -361,7 +365,7 @@ apply_key_map(keyboard_mode mode, int sym, uint32_t modifiers, char *response)
 	struct key_map map;
 	int len = 0;
 	int i = 0;
-	
+
 	while (mode[i].sym) {
 		map = mode[i++];
 		if (sym == map.sym) {
@@ -371,7 +375,7 @@ apply_key_map(keyboard_mode mode, int sym, uint32_t modifiers, char *response)
 			break;
 		}
 	}
-	
+
 	return len;
 }
 
@@ -447,6 +451,7 @@ struct terminal {
 	int width, height, row, column, max_width;
 	uint32_t buffer_height;
 	uint32_t start, end, saved_start, log_size;
+	wl_fixed_t smooth_scroll;
 	int saved_row, saved_column;
 	int scrolling;
 	int send_cursor_position;
@@ -482,7 +487,7 @@ static void
 terminal_init_tabs(struct terminal *terminal)
 {
 	int i = 0;
-	
+
 	while (i < terminal->width) {
 		if (i % 8 == 0)
 			terminal->tab_ruler[i] = 1;
@@ -653,15 +658,15 @@ terminal_scroll_window(struct terminal *terminal, int d)
 	int i;
 	int window_height;
 	int from_row, to_row;
-	
+
 	// scrolling range is inclusive
 	window_height = terminal->margin_bottom - terminal->margin_top + 1;
 	d = d % (window_height + 1);
-	if(d < 0) {
+	if (d < 0) {
 		d = 0 - d;
 		to_row = terminal->margin_bottom;
 		from_row = terminal->margin_bottom - d;
-		
+
 		for (i = 0; i < (window_height - d); i++) {
 			memcpy(terminal_get_row(terminal, to_row - i),
 			       terminal_get_row(terminal, from_row - i),
@@ -678,7 +683,7 @@ terminal_scroll_window(struct terminal *terminal, int d)
 	} else {
 		to_row = terminal->margin_top;
 		from_row = terminal->margin_top + d;
-		
+
 		for (i = 0; i < (window_height - d); i++) {
 			memcpy(terminal_get_row(terminal, to_row + i),
 			       terminal_get_row(terminal, from_row + i),
@@ -698,7 +703,7 @@ terminal_scroll_window(struct terminal *terminal, int d)
 static void
 terminal_scroll(struct terminal *terminal, int d)
 {
-	if(terminal->margin_top == 0 && terminal->margin_bottom == terminal->height - 1)
+	if (terminal->margin_top == 0 && terminal->margin_bottom == terminal->height - 1)
 		terminal_scroll_buffer(terminal, d);
 	else
 		terminal_scroll_window(terminal, d);
@@ -709,7 +714,7 @@ terminal_shift_line(struct terminal *terminal, int d)
 {
 	union utf8_char *row;
 	struct attr *attr_row;
-	
+
 	row = terminal_get_row(terminal, terminal->row);
 	attr_row = terminal_get_attr_row(terminal, terminal->row);
 
@@ -717,7 +722,7 @@ terminal_shift_line(struct terminal *terminal, int d)
 		d = terminal->column + 1 - terminal->width;
 	if ((terminal->column + d) >= terminal->width)
 		d = terminal->width - terminal->column - 1;
-	
+
 	if (d < 0) {
 		d = 0 - d;
 		memmove(&row[terminal->column],
@@ -833,13 +838,27 @@ terminal_resize_cells(struct terminal *terminal,
 	ioctl(terminal->master, TIOCSWINSZ, &ws);
 }
 
+static void
+update_title(struct terminal *terminal)
+{
+	if (window_is_resizing(terminal->window)) {
+		char *p;
+		if (asprintf(&p, "%s — [%dx%d]", terminal->title, terminal->width, terminal->height) > 0) {
+			window_set_title(terminal->window, p);
+			free(p);
+		}
+	} else {
+		window_set_title(terminal->window, terminal->title);
+	}
+}
+
 static void
 resize_handler(struct widget *widget,
 	       int32_t width, int32_t height, void *data)
 {
 	struct terminal *terminal = data;
 	int32_t columns, rows, m;
-	char *p;
+
 	m = 2 * terminal->margin;
 	columns = (width - m) / (int32_t) terminal->average_width;
 	rows = (height - m) / (int32_t) terminal->extents.height;
@@ -849,14 +868,17 @@ resize_handler(struct widget *widget,
 		width = columns * terminal->average_width + m;
 		height = rows * terminal->extents.height + m;
 		widget_set_size(terminal->widget, width, height);
-		if (asprintf(&p, "%s — [%dx%d]", terminal->title, columns, rows) > 0) {
-		    window_set_title(terminal->window, p);
-		    terminal->size_in_title = 1;
-		    free(p);
-		}
 	}
 
 	terminal_resize_cells(terminal, columns, rows);
+	update_title(terminal);
+}
+
+static void
+state_changed_handler(struct window *window, void *data)
+{
+	struct terminal *terminal = data;
+	update_title(terminal);
 }
 
 static void
@@ -922,7 +944,7 @@ terminal_send_selection(struct terminal *terminal, int fd)
 		close(fd);
 		return;
 	}
-	for (row = 0; row < terminal->height; row++) {
+	for (row = terminal->selection_start_row; row < terminal->height; row++) {
 		p_row = terminal_get_row(terminal, row);
 		for (col = 0; col < terminal->width; col++) {
 			if (p_row[col].ch == 0x200B) /* space glyph */
@@ -1177,7 +1199,7 @@ handle_term_parameter(struct terminal *terminal, int code, int sr)
 				terminal_resize(terminal, 132, 24);
 			else
 				terminal_resize(terminal, 80, 24);
-			
+
 			/* set columns, but also home cursor and clear screen */
 			terminal->row = 0; terminal->column = 0;
 			for (i = 0; i < terminal->height; i++) {
@@ -1276,6 +1298,8 @@ handle_osc(struct terminal *terminal)
 		break;
 	case 7: /* shell cwd as uri */
 		break;
+	case 777: /* Desktop notifications */
+		break;
 	default:
 		fprintf(stderr, "Unknown OSC escape code %d, text %s\n",
 			code, p);
@@ -1310,14 +1334,14 @@ handle_escape(struct terminal *terminal)
 			set[i] = 1;
 		}
 	}
-	
+
 	switch (*p) {
-	case '@':    /* ICH */
+	case '@':    /* ICH - Insert <count> blank characters */
 		count = set[0] ? args[0] : 1;
 		if (count == 0) count = 1;
 		terminal_shift_line(terminal, count);
 		break;
-	case 'A':    /* CUU */
+	case 'A':    /* CUU - Move cursor up <count> rows */
 		count = set[0] ? args[0] : 1;
 		if (count == 0) count = 1;
 		if (terminal->row - count >= terminal->margin_top)
@@ -1325,7 +1349,7 @@ handle_escape(struct terminal *terminal)
 		else
 			terminal->row = terminal->margin_top;
 		break;
-	case 'B':    /* CUD */
+	case 'B':    /* CUD - Move cursor down <count> rows */
 		count = set[0] ? args[0] : 1;
 		if (count == 0) count = 1;
 		if (terminal->row + count <= terminal->margin_bottom)
@@ -1333,7 +1357,7 @@ handle_escape(struct terminal *terminal)
 		else
 			terminal->row = terminal->margin_bottom;
 		break;
-	case 'C':    /* CUF */
+	case 'C':    /* CUF - Move cursor right by <count> columns */
 		count = set[0] ? args[0] : 1;
 		if (count == 0) count = 1;
 		if ((terminal->column + count) < terminal->width)
@@ -1341,7 +1365,7 @@ handle_escape(struct terminal *terminal)
 		else
 			terminal->column = terminal->width - 1;
 		break;
-	case 'D':    /* CUB */
+	case 'D':    /* CUB - Move cursor left <count> columns */
 		count = set[0] ? args[0] : 1;
 		if (count == 0) count = 1;
 		if ((terminal->column - count) >= 0)
@@ -1349,7 +1373,7 @@ handle_escape(struct terminal *terminal)
 		else
 			terminal->column = 0;
 		break;
-	case 'E':    /* CNL */
+	case 'E':    /* CNL - Move cursor down <count> rows, to column 1 */
 		count = set[0] ? args[0] : 1;
 		if (terminal->row + count <= terminal->margin_bottom)
 			terminal->row += count;
@@ -1357,7 +1381,7 @@ handle_escape(struct terminal *terminal)
 			terminal->row = terminal->margin_bottom;
 		terminal->column = 0;
 		break;
-	case 'F':    /* CPL */
+	case 'F':    /* CPL - Move cursour up <count> rows, to column 1 */
 		count = set[0] ? args[0] : 1;
 		if (terminal->row - count >= terminal->margin_top)
 			terminal->row -= count;
@@ -1365,18 +1389,18 @@ handle_escape(struct terminal *terminal)
 			terminal->row = terminal->margin_top;
 		terminal->column = 0;
 		break;
-	case 'G':    /* CHA */
+	case 'G':    /* CHA - Move cursor to column <y> in current row */
 		y = set[0] ? args[0] : 1;
 		y = y <= 0 ? 1 : y > terminal->width ? terminal->width : y;
-		
+
 		terminal->column = y - 1;
 		break;
-	case 'f':    /* HVP */
-	case 'H':    /* CUP */
+	case 'f':    /* HVP - Move cursor to <x, y> */
+	case 'H':    /* CUP - Move cursor to <x, y> (origin at 1,1) */
 		x = (set[1] ? args[1] : 1) - 1;
 		x = x < 0 ? 0 :
 		    (x >= terminal->width ? terminal->width - 1 : x);
-		
+
 		y = (set[0] ? args[0] : 1) - 1;
 		if (terminal->origin_mode) {
 			y += terminal->margin_top;
@@ -1386,7 +1410,7 @@ handle_escape(struct terminal *terminal)
 			y = y < 0 ? 0 :
 			    (y >= terminal->height ? terminal->height - 1 : y);
 		}
-		
+
 		terminal->row = y;
 		terminal->column = x;
 		break;
@@ -1399,7 +1423,7 @@ handle_escape(struct terminal *terminal)
 		}
 		terminal->column--;
 		break;
-	case 'J':    /* ED */
+	case 'J':    /* ED - Erase display */
 		row = terminal_get_row(terminal, terminal->row);
 		attr_row = terminal_get_attr_row(terminal, terminal->row);
 		if (!set[0] || args[0] == 0 || args[0] > 2) {
@@ -1428,7 +1452,7 @@ handle_escape(struct terminal *terminal)
 					       terminal->end - terminal->start);
 		}
 		break;
-	case 'K':    /* EL */
+	case 'K':    /* EL - Erase line */
 		row = terminal_get_row(terminal, terminal->row);
 		attr_row = terminal_get_attr_row(terminal, terminal->row);
 		if (!set[0] || args[0] == 0 || args[0] > 2) {
@@ -1444,7 +1468,7 @@ handle_escape(struct terminal *terminal)
 			attr_init(attr_row, terminal->curr_attr, terminal->width);
 		}
 		break;
-	case 'L':    /* IL */
+	case 'L':    /* IL - Insert <count> blank lines */
 		count = set[0] ? args[0] : 1;
 		if (count == 0) count = 1;
 		if (terminal->row >= terminal->margin_top &&
@@ -1461,7 +1485,7 @@ handle_escape(struct terminal *terminal)
 				terminal->curr_attr, terminal->width);
 		}
 		break;
-	case 'M':    /* DL */
+	case 'M':    /* DL - Delete <count> lines */
 		count = set[0] ? args[0] : 1;
 		if (count == 0) count = 1;
 		if (terminal->row >= terminal->margin_top &&
@@ -1476,7 +1500,7 @@ handle_escape(struct terminal *terminal)
 			       0, terminal->data_pitch);
 		}
 		break;
-	case 'P':    /* DCH */
+	case 'P':    /* DCH - Delete <count> characters on current line */
 		count = set[0] ? args[0] : 1;
 		if (count == 0) count = 1;
 		terminal_shift_line(terminal, 0 - count);
@@ -1487,7 +1511,7 @@ handle_escape(struct terminal *terminal)
 	case 'T':    /* SD */
 		terminal_scroll(terminal, 0 - (set[0] ? args[0] : 1));
 		break;
-	case 'X':    /* ECH */
+	case 'X':    /* ECH - Erase <count> characters on current line */
 		count = set[0] ? args[0] : 1;
 		if (count == 0) count = 1;
 		if ((terminal->column + count) > terminal->width)
@@ -1506,10 +1530,10 @@ handle_escape(struct terminal *terminal)
 		}
 		terminal->column++;
 		break;
-	case '`':    /* HPA */
+	case '`':    /* HPA - Move cursor to <y> column in current row */
 		y = set[0] ? args[0] : 1;
 		y = y <= 0 ? 1 : y > terminal->width ? terminal->width : y;
-		
+
 		terminal->column = y - 1;
 		break;
 	case 'b':    /* REP */
@@ -1520,34 +1544,34 @@ handle_escape(struct terminal *terminal)
 				handle_char(terminal, terminal->last_char);
 		terminal->last_char.byte[0] = 0;
 		break;
-	case 'c':    /* Primary DA */
+	case 'c':    /* Primary DA - Answer "I am a VT102" */
 		terminal_write(terminal, "\e[?6c", 5);
 		break;
-	case 'd':    /* VPA */
+	case 'd':    /* VPA - Move cursor to <x> row, current column */
 		x = set[0] ? args[0] : 1;
 		x = x <= 0 ? 1 : x > terminal->height ? terminal->height : x;
-		
+
 		terminal->row = x - 1;
 		break;
-	case 'g':    /* TBC */
+	case 'g':    /* TBC - Clear tab stop(s) */
 		if (!set[0] || args[0] == 0) {
 			terminal->tab_ruler[terminal->column] = 0;
 		} else if (args[0] == 3) {
 			memset(terminal->tab_ruler, 0, terminal->width);
 		}
 		break;
-	case 'h':    /* SM */
-		for(i = 0; i < 10 && set[i]; i++) {
+	case 'h':    /* SM - Set mode */
+		for (i = 0; i < 10 && set[i]; i++) {
 			handle_term_parameter(terminal, args[i], 1);
 		}
 		break;
-	case 'l':    /* RM */
-		for(i = 0; i < 10 && set[i]; i++) {
+	case 'l':    /* RM - Reset mode */
+		for (i = 0; i < 10 && set[i]; i++) {
 			handle_term_parameter(terminal, args[i], 0);
 		}
 		break;
-	case 'm':    /* SGR */
-		for(i = 0; i < 10; i++) {
+	case 'm':    /* SGR - Set attributes */
+		for (i = 0; i < 10; i++) {
 			if (i <= 7 && set[i] && set[i + 1] &&
 				set[i + 2] && args[i + 1] == 5)
 			{
@@ -1559,9 +1583,9 @@ handle_escape(struct terminal *terminal)
 					break;
 				}
 			}
-			if(set[i]) {
+			if (set[i]) {
 				handle_sgr(terminal, args[i]);
-			} else if(i == 0) {
+			} else if (i == 0) {
 				handle_sgr(terminal, 0);
 				break;
 			} else {
@@ -1569,7 +1593,7 @@ handle_escape(struct terminal *terminal)
 			}
 		}
 		break;
-	case 'n':    /* DSR */
+	case 'n':    /* DSR - Status report */
 		i = set[0] ? args[0] : 0;
 		if (i == 0 || i == 5) {
 			terminal_write(terminal, "\e[0n", 4);
@@ -1581,8 +1605,8 @@ handle_escape(struct terminal *terminal)
 			terminal_write(terminal, response, strlen(response));
 		}
  		break;
-	case 'r':
-		if(!set[0]) {
+	case 'r':    /* DECSTBM - Set scrolling region */
+		if (!set[0]) {
 			terminal->margin_top = 0;
 			terminal->margin_bottom = terminal->height-1;
 			terminal->row = 0;
@@ -1594,21 +1618,21 @@ handle_escape(struct terminal *terminal)
 			bottom = (set[1] ? args[1] : 1) - 1;
 			bottom = bottom < 0 ? 0 :
 			         (bottom >= terminal->height ? terminal->height - 1 : bottom);
-			if(bottom > top) {
+			if (bottom > top) {
 				terminal->margin_top = top;
 				terminal->margin_bottom = bottom;
 			} else {
 				terminal->margin_top = 0;
 				terminal->margin_bottom = terminal->height-1;
 			}
-			if(terminal->origin_mode)
+			if (terminal->origin_mode)
 				terminal->row = terminal->margin_top;
 			else
 				terminal->row = 0;
 			terminal->column = 0;
 		}
 		break;
-	case 's':
+	case 's':    /* Save cursor location */
 		terminal->saved_row = terminal->row;
 		terminal->saved_column = terminal->column;
 		break;
@@ -1655,44 +1679,44 @@ handle_escape(struct terminal *terminal)
 				fprintf(stderr, "Unimplemented windowOp %d\n", args[0]);
 			break;
 		}
-	case 'u':
+	case 'u':    /* Restore cursor location */
 		terminal->row = terminal->saved_row;
 		terminal->column = terminal->saved_column;
 		break;
 	default:
 		fprintf(stderr, "Unknown CSI escape: %c\n", *p);
 		break;
-	}	
+	}
 }
 
 static void
 handle_non_csi_escape(struct terminal *terminal, char code)
 {
 	switch(code) {
-	case 'M':    /* RI */
+	case 'M':    /* RI - Reverse linefeed */
 		terminal->row -= 1;
-		if(terminal->row < terminal->margin_top) {
+		if (terminal->row < terminal->margin_top) {
 			terminal->row = terminal->margin_top;
 			terminal_scroll(terminal, -1);
 		}
 		break;
-	case 'E':    /* NEL */
+	case 'E':    /* NEL - Newline */
 		terminal->column = 0;
 		// fallthrough
-	case 'D':    /* IND */
+	case 'D':    /* IND - Linefeed */
 		terminal->row += 1;
-		if(terminal->row > terminal->margin_bottom) {
+		if (terminal->row > terminal->margin_bottom) {
 			terminal->row = terminal->margin_bottom;
 			terminal_scroll(terminal, +1);
 		}
 		break;
-	case 'c':    /* RIS */
+	case 'c':    /* RIS - Reset*/
 		terminal_init(terminal);
 		break;
-	case 'H':    /* HTS */
+	case 'H':    /* HTS - Set tab stop at current column */
 		terminal->tab_ruler[terminal->column] = 1;
 		break;
-	case '7':    /* DECSC */
+	case '7':    /* DECSC - Save current state */
 		terminal->saved_row = terminal->row;
 		terminal->saved_column = terminal->column;
 		terminal->saved_attr = terminal->curr_attr;
@@ -1701,7 +1725,7 @@ handle_non_csi_escape(struct terminal *terminal, char code)
 		terminal->saved_g0 = terminal->g0;
 		terminal->saved_g1 = terminal->g1;
 		break;
-	case '8':    /* DECRC */
+	case '8':    /* DECRC - Restore state most recently saved by ESC 7 */
 		terminal->row = terminal->saved_row;
 		terminal->column = terminal->saved_column;
 		terminal->curr_attr = terminal->saved_attr;
@@ -1710,10 +1734,10 @@ handle_non_csi_escape(struct terminal *terminal, char code)
 		terminal->g0 = terminal->saved_g0;
 		terminal->g1 = terminal->saved_g1;
 		break;
-	case '=':    /* DECPAM */
+	case '=':    /* DECPAM - Set application keypad mode */
 		terminal->key_mode = KM_APPLICATION;
 		break;
-	case '>':    /* DECPNM */
+	case '>':    /* DECPNM - Set numeric keypad mode */
 		terminal->key_mode = KM_NORMAL;
 		break;
 	default:
@@ -1733,7 +1757,7 @@ handle_special_escape(struct terminal *terminal, char special, char code)
 			/* fill with 'E', no cheap way to do this */
 			memset(terminal->data, 0, terminal->data_pitch * terminal->height);
 			numChars = terminal->width * terminal->height;
-			for(i = 0; i < numChars; i++) {
+			for (i = 0; i < numChars; i++) {
 				terminal->data[i].byte[0] = 'E';
 			}
 			break;
@@ -1821,19 +1845,19 @@ handle_sgr(struct terminal *terminal, int code)
 		terminal->curr_attr.bg = terminal->color_scheme->default_attr.bg;
 		break;
 	default:
-		if(code >= 30 && code <= 37) {
+		if (code >= 30 && code <= 37) {
 			terminal->curr_attr.fg = code - 30;
 			if (terminal->curr_attr.a & ATTRMASK_BOLD)
 				terminal->curr_attr.fg += 8;
-		} else if(code >= 40 && code <= 47) {
+		} else if (code >= 40 && code <= 47) {
 			terminal->curr_attr.bg = code - 40;
 		} else if (code >= 90 && code <= 97) {
 			terminal->curr_attr.fg = code - 90 + 8;
 		} else if (code >= 100 && code <= 107) {
 			terminal->curr_attr.bg = code - 100 + 8;
-		} else if(code >= 256 && code < 512) {
+		} else if (code >= 256 && code < 512) {
 			terminal->curr_attr.fg = code - 256;
-		} else if(code >= 512 && code < 768) {
+		} else if (code >= 512 && code < 768) {
 			terminal->curr_attr.bg = code - 512;
 		} else {
 			fprintf(stderr, "Unknown SGR code: %d\n", code);
@@ -1918,7 +1942,7 @@ handle_special_char(struct terminal *terminal, char c)
 	default:
 		return 0;
 	}
-	
+
 	return 1;
 }
 
@@ -1927,24 +1951,24 @@ handle_char(struct terminal *terminal, union utf8_char utf8)
 {
 	union utf8_char *row;
 	struct attr *attr_row;
-	
+
 	if (handle_special_char(terminal, utf8.byte[0])) return;
 
 	apply_char_set(terminal->cs, &utf8);
-	
+
 	/* There are a whole lot of non-characters, control codes,
 	 * and formatting codes that should probably be ignored,
 	 * for example: */
 	if (strncmp((char*) utf8.byte, "\xEF\xBB\xBF", 3) == 0) {
 		/* BOM, ignore */
 		return;
-	} 
-	
+	}
+
 	/* Some of these non-characters should be translated, e.g.: */
 	if (utf8.byte[0] < 32) {
 		utf8.byte[0] = utf8.byte[0] + 64;
 	}
-	
+
 	/* handle right margin effects */
 	if (terminal->column >= terminal->width) {
 		if (terminal->mode & MODE_AUTOWRAP) {
@@ -1958,10 +1982,10 @@ handle_char(struct terminal *terminal, union utf8_char utf8)
 			terminal->column--;
  		}
  	}
-	
+
 	row = terminal_get_row(terminal, terminal->row);
 	attr_row = terminal_get_attr_row(terminal, terminal->row);
-	
+
 	if (terminal->mode & MODE_IRM)
 		terminal_shift_line(terminal, +1);
 	row[terminal->column] = utf8;
@@ -2078,7 +2102,7 @@ terminal_data(struct terminal *terminal, const char *data, size_t length)
 				if (terminal->escape_length >= MAX_ESCAPE)
 					terminal->state = escape_state_normal;
 			}
-			
+
 			if (isalpha(utf8.byte[0]) || utf8.byte[0] == '@' ||
 				utf8.byte[0] == '`')
 			{
@@ -2485,7 +2509,7 @@ key_handler(struct window *window, struct input *input, uint32_t time,
 		/* Handle special keys with alternate mappings */
 		len = apply_key_map(terminal->key_mode, sym, modifiers, ch);
 		if (len != 0) break;
-		
+
 		if (modifiers & MOD_CONTROL_MASK) {
 			if (sym >= '3' && sym <= '7')
 				sym = (sym & 0x1f) + 8;
@@ -2559,7 +2583,7 @@ static int wordsep(int ch)
 {
 	const char extra[] = "-,./?%&#:_=+@~";
 
-	if (ch > 127)
+	if (ch > 127 || ch < 0)
 		return 1;
 
 	return ch == 0 || !(isalpha(ch) || isdigit(ch) || strchr(extra, ch));
@@ -2663,9 +2687,16 @@ recompute_selection(struct terminal *terminal)
 }
 
 static void
-menu_func(struct window *window, struct input *input, int index, void *data)
+terminal_minimize(struct terminal *terminal)
 {
-	struct terminal *terminal = data;
+	window_set_minimized(terminal->window);
+}
+
+static void
+menu_func(void *data, struct input *input, int index)
+{
+	struct window *window = data;
+	struct terminal *terminal = window_get_user_data(window);
 
 	fprintf(stderr, "picked entry %d\n", index);
 
@@ -2679,6 +2710,9 @@ menu_func(struct window *window, struct input *input, int index, void *data)
 	case 2:
 		terminal_paste(terminal, input);
 		break;
+	case 3:
+		terminal_minimize(terminal);
+		break;
 	}
 }
 
@@ -2687,7 +2721,7 @@ show_menu(struct terminal *terminal, struct input *input, uint32_t time)
 {
 	int32_t x, y;
 	static const char *entries[] = {
-		"Open Terminal", "Copy", "Paste"
+		"Open Terminal", "Copy", "Paste", "Minimize"
 	};
 
 	input_get_position(input, &x, &y);
@@ -2745,14 +2779,6 @@ static int
 enter_handler(struct widget *widget,
 	      struct input *input, float x, float y, void *data)
 {
-	struct terminal *terminal = data;
-
-	/* Reset title to get rid of resizing '[WxH]' in titlebar */
-	if (terminal->size_in_title) {
-		window_set_title(terminal->window, terminal->title);
-		terminal->size_in_title = 0;
-	}
-
 	return CURSOR_IBEAM;
 }
 
@@ -2775,6 +2801,55 @@ motion_handler(struct widget *widget,
 	return CURSOR_IBEAM;
 }
 
+/* This magnitude is chosen rather arbitrarily. Really, the scrolling
+ * should happen on a (fractional) pixel basis, not a line basis. */
+#define AXIS_UNITS_PER_LINE 256
+
+static void
+axis_handler(struct widget *widget,
+	     struct input *input, uint32_t time,
+	     uint32_t axis,
+	     wl_fixed_t value,
+	     void *data)
+{
+	struct terminal *terminal = data;
+	int lines;
+
+	if (axis != WL_POINTER_AXIS_VERTICAL_SCROLL)
+		return;
+
+	terminal->smooth_scroll += value;
+	lines = terminal->smooth_scroll / AXIS_UNITS_PER_LINE;
+	terminal->smooth_scroll -= lines * AXIS_UNITS_PER_LINE;
+
+	if (lines > 0) {
+		if (terminal->scrolling) {
+			if ((uint32_t)lines > terminal->saved_start - terminal->start)
+				lines = terminal->saved_start - terminal->start;
+		} else {
+			lines = 0;
+		}
+	} else if (lines < 0) {
+		uint32_t neg_lines = -lines;
+
+		if (neg_lines > terminal->log_size + terminal->start - terminal->end)
+			lines = terminal->end - terminal->log_size - terminal->start;
+	}
+
+	if (lines) {
+		if (!terminal->scrolling)
+			terminal->saved_start = terminal->start;
+		terminal->scrolling = 1;
+
+		terminal->start += lines;
+		terminal->row -= lines;
+		terminal->selection_start_row -= lines;
+		terminal->selection_end_row -= lines;
+
+		widget_schedule_redraw(widget);
+	}
+}
+
 static void
 output_handler(struct window *window, struct output *output, int enter,
 	       void *data)
@@ -2860,6 +2935,7 @@ terminal_create(struct display *display)
 	window_set_fullscreen_handler(terminal->window, fullscreen_handler);
 	window_set_output_handler(terminal->window, output_handler);
 	window_set_close_handler(terminal->window, close_handler);
+	window_set_state_changed_handler(terminal->window, state_changed_handler);
 
 	window_set_data_handler(terminal->window, data_handler);
 	window_set_drop_handler(terminal->window, drop_handler);
@@ -2869,6 +2945,7 @@ terminal_create(struct display *display)
 	widget_set_button_handler(terminal->widget, button_handler);
 	widget_set_enter_handler(terminal->widget, enter_handler);
 	widget_set_motion_handler(terminal->widget, motion_handler);
+	widget_set_axis_handler(terminal->widget, axis_handler);
 	widget_set_touch_up_handler(terminal->widget, touch_up_handler);
 	widget_set_touch_down_handler(terminal->widget, touch_down_handler);
 	widget_set_touch_motion_handler(terminal->widget, touch_motion_handler);
@@ -2980,6 +3057,7 @@ terminal_run(struct terminal *terminal, const char *path)
 static const struct weston_option terminal_options[] = {
 	{ WESTON_OPTION_BOOLEAN, "fullscreen", 'f', &option_fullscreen },
 	{ WESTON_OPTION_STRING, "font", 0, &option_font },
+	{ WESTON_OPTION_INTEGER, "font-size", 0, &option_font_size },
 	{ WESTON_OPTION_STRING, "shell", 0, &option_shell },
 };
 
@@ -2987,6 +3065,7 @@ int main(int argc, char *argv[])
 {
 	struct display *d;
 	struct terminal *terminal;
+	const char *config_file;
 	struct weston_config *config;
 	struct weston_config_section *s;
 
@@ -2998,13 +3077,24 @@ int main(int argc, char *argv[])
 	if (!option_shell)
 		option_shell = "/bin/bash";
 
-	config = weston_config_parse("weston.ini");
+	config_file = weston_config_get_name_from_env();
+	config = weston_config_parse(config_file);
 	s = weston_config_get_section(config, "terminal", NULL, NULL);
 	weston_config_section_get_string(s, "font", &option_font, "mono");
 	weston_config_section_get_int(s, "font-size", &option_font_size, 14);
 	weston_config_section_get_string(s, "term", &option_term, "xterm");
 	weston_config_destroy(config);
 
+	if (parse_options(terminal_options,
+			  ARRAY_LENGTH(terminal_options), &argc, argv) > 1) {
+		printf("Usage: %s [OPTIONS]\n"
+		       "  --fullscreen or -f\n"
+		       "  --font=NAME\n"
+		       "  --font-size=SIZE\n"
+		       "  --shell=NAME\n", argv[0]);
+		return 1;
+	}
+
 	d = display_create(&argc, argv);
 	if (d == NULL) {
 		fprintf(stderr, "failed to create display: %m\n");
