@@ -2,37 +2,44 @@
  * Copyright © 2008-2012 Kristian Høgsberg
  * Copyright © 2012 Intel Corporation
  *
- * Permission to use, copy, modify, distribute, and sell this software and its
- * documentation for any purpose is hereby granted without fee, provided that
- * the above copyright notice appear in all copies and that both that copyright
- * notice and this permission notice appear in supporting documentation, and
- * that the name of the copyright holders not be used in advertising or
- * publicity pertaining to distribution of the software without specific,
- * written prior permission.  The copyright holders make no representations
- * about the suitability of this software for any purpose.  It is provided "as
- * is" without express or implied warranty.
+ * Permission is hereby granted, free of charge, to any person obtaining
+ * a copy of this software and associated documentation files (the
+ * "Software"), to deal in the Software without restriction, including
+ * without limitation the rights to use, copy, modify, merge, publish,
+ * distribute, sublicense, and/or sell copies of the Software, and to
+ * permit persons to whom the Software is furnished to do so, subject to
+ * the following conditions:
  *
- * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
- * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
- * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
- * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
- * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
- * OF THIS SOFTWARE.
+ * The above copyright notice and this permission notice (including the
+ * next paragraph) shall be included in all copies or substantial
+ * portions of the Software.
+ *
+ * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
+ * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
+ * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
+ * NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
+ * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
+ * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
+ * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
+ * SOFTWARE.
  */
 
 #include "config.h"
 
+#include <errno.h>
 #include <stdlib.h>
+#include <stdint.h>
 #include <stdio.h>
 #include <string.h>
-#include <jpeglib.h>
 #include <png.h>
 #include <pixman.h>
 
+#include "shared/helpers.h"
 #include "image-loader.h"
 
-#define ARRAY_LENGTH(a) (sizeof (a) / sizeof (a)[0])
+#ifdef HAVE_JPEG
+#include <jpeglib.h>
+#endif
 
 #ifdef HAVE_WEBP
 #include <webp/decode.h>
@@ -44,6 +51,14 @@ stride_for_width(int width)
 	return width * 4;
 }
 
+static void
+pixman_image_destroy_func(pixman_image_t *image, void *data)
+{
+	free(data);
+}
+
+#ifdef HAVE_JPEG
+
 static void
 swizzle_row(JSAMPLE *row, JDIMENSION width)
 {
@@ -65,12 +80,6 @@ error_exit(j_common_ptr cinfo)
 	longjmp(cinfo->client_data, 1);
 }
 
-static void
-pixman_image_destroy_func(pixman_image_t *image, void *data)
-{
-	free(data);
-}
-
 static pixman_image_t *
 load_jpeg(FILE *fp)
 {
@@ -129,6 +138,17 @@ load_jpeg(FILE *fp)
 	return pixman_image;
 }
 
+#else
+
+static pixman_image_t *
+load_jpeg(FILE *fp)
+{
+	fprintf(stderr, "JPEG support disabled at compile-time\n");
+	return NULL;
+}
+
+#endif
+
 static inline int
 multiply_alpha(int alpha, int color)
 {
@@ -349,6 +369,15 @@ load_webp(FILE *fp)
 					config.output.u.RGBA.stride);
 }
 
+#else
+
+static pixman_image_t *
+load_webp(FILE *fp)
+{
+	fprintf(stderr, "WebP support disabled at compile-time\n");
+	return NULL;
+}
+
 #endif
 
 
@@ -361,25 +390,29 @@ struct image_loader {
 static const struct image_loader loaders[] = {
 	{ { 0x89, 'P', 'N', 'G' }, 4, load_png },
 	{ { 0xff, 0xd8 }, 2, load_jpeg },
-#ifdef HAVE_WEBP
 	{ { 'R', 'I', 'F', 'F' }, 4, load_webp }
-#endif
 };
 
 pixman_image_t *
 load_image(const char *filename)
 {
-	pixman_image_t *image;
+	pixman_image_t *image = NULL;
 	unsigned char header[4];
 	FILE *fp;
 	unsigned int i;
 
-	fp = fopen(filename, "rb");
-	if (fp == NULL)
+	if (!filename || !*filename)
 		return NULL;
 
+	fp = fopen(filename, "rb");
+	if (!fp) {
+		fprintf(stderr, "%s: %s\n", filename, strerror(errno));
+		return NULL;
+	}
+
 	if (fread(header, sizeof header, 1, fp) != 1) {
 		fclose(fp);
+		fprintf(stderr, "%s: unable to read file header\n", filename);
 		return NULL;
 	}
 
@@ -395,10 +428,12 @@ load_image(const char *filename)
 	fclose(fp);
 
 	if (i == ARRAY_LENGTH(loaders)) {
-		fprintf(stderr, "unrecognized file header for %s: "
+		fprintf(stderr, "%s: unrecognized file header "
 			"0x%02x 0x%02x 0x%02x 0x%02x\n",
 			filename, header[0], header[1], header[2], header[3]);
-		image = NULL;
+	} else if (!image) {
+		/* load probably printed something, but just in case */
+		fprintf(stderr, "%s: error reading image\n", filename);
 	}
 
 	return image;
